@@ -30,10 +30,6 @@
  * SUCH DAMAGE.
  */
 
-/* Enable mprotect protection munging.  ARM and MIPS Linux needs this
- * it seems, so leave this enabled by default */
-#define DO_MPROTECT_HACKS
-
 // Support a list of library preloads in /etc/ld.so.preload
 //#define SUPPORT_LDSO_PRELOAD_FILE
 
@@ -109,12 +105,8 @@
  * application.
  */
 
-#include "ld_syscall.h"
-#include "linuxelf.h"
-#include "ld_hash.h"
-#include "ld_string.h"
-#include "dlfcn.h"
-#include "../config.h"
+#include "ldso.h"
+
 
 #define ALLOW_ZERO_PLTGOT
 
@@ -133,16 +125,16 @@
 char *_dl_library_path = 0;		/* Where we look for libraries */
 char *_dl_preload = 0;			/* Things to be loaded before the libs. */
 char *_dl_ldsopath = 0;
-static char *_dl_not_lazy = 0;
+static int _dl_be_lazy = RTLD_LAZY;
 #ifdef __SUPPORT_LD_DEBUG__
-static char *_dl_debug  = 0;
-static char *_dl_debug_symbols = 0;
-static char *_dl_debug_move    = 0;
-static char *_dl_debug_reloc   = 0;
-static char *_dl_debug_detail  = 0;
-static char *_dl_debug_nofixups  = 0;
-static char *_dl_debug_bindings  = 0;
-static int   _dl_debug_file = 2;
+char *_dl_debug  = 0;
+char *_dl_debug_symbols = 0;
+char *_dl_debug_move    = 0;
+char *_dl_debug_reloc   = 0;
+char *_dl_debug_detail  = 0;
+char *_dl_debug_nofixups  = 0;
+char *_dl_debug_bindings  = 0;
+int   _dl_debug_file = 2;
 #else
 #define _dl_debug_file 2
 #endif
@@ -150,20 +142,18 @@ static char *_dl_malloc_addr, *_dl_mmap_zero;
 
 static char *_dl_trace_loaded_objects = 0;
 static int (*_dl_elf_main) (int, char **, char **);
-static int (*_dl_elf_init) (void);
-void *(*_dl_malloc_function) (int size) = NULL;
 struct r_debug *_dl_debug_addr = NULL;
 unsigned long *_dl_brkp;
 unsigned long *_dl_envp;
-int _dl_fixup(struct elf_resolve *tpnt);
+int _dl_fixup(struct elf_resolve *tpnt, int lazy);
 void _dl_debug_state(void);
 char *_dl_get_last_path_component(char *path);
 static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *app_tpnt, 
 		unsigned long load_addr, unsigned long *hash_addr, Elf32_auxv_t auxvt[AT_EGID + 1], 
 		char **envp, struct r_debug *debug_addr);
-#include "boot1_arch.h"
-#include "ldso.h"				/* Pull in the name of ld.so */
 
+#include "boot1_arch.h"
+#include "_dl_progname.h"				/* Pull in the value of _dl_progname */
 
 /* When we enter this piece of code, the program stack looks like this:
         argc            argument counter (integer)
@@ -207,15 +197,14 @@ LD_BOOT(unsigned long args)
 	unsigned long *got;
 	unsigned long *aux_dat;
 	int goof = 0;
-	elfhdr *header;
+	ElfW(Ehdr) *header;
 	struct elf_resolve *tpnt;
 	struct elf_resolve *app_tpnt;
 	Elf32_auxv_t auxvt[AT_EGID + 1];
 	unsigned char *malloc_buffer, *mmap_zero;
 	Elf32_Dyn *dpnt;
-	Elf32_Dyn *dpnt_debug = NULL;
 	unsigned long *hash_addr;
-	struct r_debug *debug_addr;
+	struct r_debug *debug_addr = NULL;
 	int indx;
 	int status;
 
@@ -228,7 +217,7 @@ LD_BOOT(unsigned long args)
 	/* First obtain the information on the stack that tells us more about
 	   what binary is loaded, where it is loaded, etc, etc */
 	GET_ARGV(aux_dat, args);
-#if defined (__arm__) || defined (__mips__)
+#if defined (__arm__) || defined (__mips__) || defined (__cris__)
 	aux_dat += 1;
 #endif
 	argc = *(aux_dat - 1);
@@ -259,7 +248,7 @@ LD_BOOT(unsigned long args)
 	/* locate the ELF header.   We need this done as soon as possible 
 	 * (esp since SEND_STDERR() needs this on some platforms... */
 	load_addr = auxvt[AT_BASE].a_un.a_val;
-	header = (elfhdr *) auxvt[AT_BASE].a_un.a_ptr;
+	header = (ElfW(Ehdr) *) auxvt[AT_BASE].a_un.a_ptr;
 
 	/* Check the ELF header to make sure everything looks ok.  */
 	if (!header || header->e_ident[EI_CLASS] != ELFCLASS32 ||
@@ -378,18 +367,21 @@ LD_BOOT(unsigned long args)
 	}
 
 	tpnt = LD_MALLOC(sizeof(struct elf_resolve));
-	_dl_memset(tpnt, 0, sizeof(*tpnt));
+	_dl_memset(tpnt, 0, sizeof(struct elf_resolve));
 	app_tpnt = LD_MALLOC(sizeof(struct elf_resolve));
-	_dl_memset(app_tpnt, 0, sizeof(*app_tpnt));
+	_dl_memset(app_tpnt, 0, sizeof(struct elf_resolve));
 
 	/*
 	 * This is used by gdb to locate the chain of shared libraries that are currently loaded.
 	 */
 	debug_addr = LD_MALLOC(sizeof(struct r_debug));
-	_dl_memset(debug_addr, 0, sizeof(*debug_addr));
+	_dl_memset(debug_addr, 0, sizeof(struct r_debug));
 
 	/* OK, that was easy.  Next scan the DYNAMIC section of the image.
 	   We are only doing ourself right now - we will have to do the rest later */
+#ifdef __SUPPORT_LD_DEBUG_EARLY__
+	SEND_STDERR("scanning DYNAMIC section\n");
+#endif
 	while (dpnt->d_tag) {
 #if defined(__mips__)
 		if (dpnt->d_tag == DT_MIPS_GOTSYM)
@@ -401,7 +393,7 @@ LD_BOOT(unsigned long args)
 #endif
 		if (dpnt->d_tag < 24) {
 			tpnt->dynamic_info[dpnt->d_tag] = dpnt->d_un.d_val;
-			if (dpnt->d_tag == DT_TEXTREL || SVR4_BUGCOMPAT) {
+			if (dpnt->d_tag == DT_TEXTREL) {
 				tpnt->dynamic_info[DT_TEXTREL] = 1;
 			}
 		}
@@ -409,10 +401,10 @@ LD_BOOT(unsigned long args)
 	}
 
 	{
-		elf_phdr *ppnt;
+		ElfW(Phdr) *ppnt;
 		int i;
 
-		ppnt = (elf_phdr *) auxvt[AT_PHDR].a_un.a_ptr;
+		ppnt = (ElfW(Phdr) *) auxvt[AT_PHDR].a_un.a_ptr;
 		for (i = 0; i < auxvt[AT_PHNUM].a_un.a_val; i++, ppnt++)
 			if (ppnt->p_type == PT_DYNAMIC) {
 				dpnt = (Elf32_Dyn *) ppnt->p_vaddr;
@@ -427,25 +419,40 @@ LD_BOOT(unsigned long args)
 					if (dpnt->d_tag == DT_MIPS_SYMTABNO)
 						app_tpnt->mips_symtabno =
 							(unsigned long) dpnt->d_un.d_val;
-#endif
 					if (dpnt->d_tag > DT_JMPREL) {
 						dpnt++;
 						continue;
 					}
 					app_tpnt->dynamic_info[dpnt->d_tag] = dpnt->d_un.d_val;
-					if (dpnt->d_tag == DT_DEBUG)
-#ifndef DO_MPROTECT_HACKS
+
+#warning "Debugging threads on mips won't work till someone fixes this..."
+#if 0
+					if (dpnt->d_tag == DT_DEBUG) {
 						dpnt->d_un.d_val = (unsigned long) debug_addr;
-#else
-						dpnt_debug = dpnt;
+					}
 #endif
-					if (dpnt->d_tag == DT_TEXTREL || SVR4_BUGCOMPAT)
+
+#else
+					if (dpnt->d_tag > DT_JMPREL) {
+						dpnt++;
+						continue;
+					}
+					app_tpnt->dynamic_info[dpnt->d_tag] = dpnt->d_un.d_val;
+					if (dpnt->d_tag == DT_DEBUG) {
+						dpnt->d_un.d_val = (unsigned long) debug_addr;
+					}
+#endif
+					if (dpnt->d_tag == DT_TEXTREL)
 						app_tpnt->dynamic_info[DT_TEXTREL] = 1;
 					dpnt++;
 				}
 			}
 	}
 
+#ifdef __SUPPORT_LD_DEBUG_EARLY__
+	SEND_STDERR("done scanning DYNAMIC section\n");
+#endif
+
 	/* Get some more of the information that we will need to dynamicly link
 	   this module to itself */
 
@@ -455,19 +462,27 @@ LD_BOOT(unsigned long args)
 	tpnt->elf_buckets = hash_addr;
 	hash_addr += tpnt->nbucket;
 
-#ifdef DO_MPROTECT_HACKS
+#ifdef __SUPPORT_LD_DEBUG_EARLY__
+	SEND_STDERR("done grabbing link information\n");
+#endif
+
+#ifndef FORCE_SHAREABLE_TEXT_SEGMENTS
 	/* Ugly, ugly.  We need to call mprotect to change the protection of
 	   the text pages so that we can do the dynamic linking.  We can set the
 	   protection back again once we are done */
 
 	{
-		elf_phdr *ppnt;
+		ElfW(Phdr) *ppnt;
 		int i;
 
+#ifdef __SUPPORT_LD_DEBUG_EARLY__
+		SEND_STDERR("calling mprotect on the shared library/dynamic linker\n");
+#endif
+
 		/* First cover the shared library/dynamic linker. */
 		if (tpnt->dynamic_info[DT_TEXTREL]) {
-			header = (elfhdr *) auxvt[AT_BASE].a_un.a_ptr;
-			ppnt = (elf_phdr *) ((int)auxvt[AT_BASE].a_un.a_ptr + 
+			header = (ElfW(Ehdr) *) auxvt[AT_BASE].a_un.a_ptr;
+			ppnt = (ElfW(Phdr) *) ((int)auxvt[AT_BASE].a_un.a_ptr + 
 					header->e_phoff);
 			for (i = 0; i < header->e_phnum; i++, ppnt++) {
 				if (ppnt->p_type == PT_LOAD && !(ppnt->p_flags & PF_W)) {
@@ -478,9 +493,12 @@ LD_BOOT(unsigned long args)
 			}
 		}
 
+#ifdef __SUPPORT_LD_DEBUG_EARLY__
+		SEND_STDERR("calling mprotect on the application program\n");
+#endif
 		/* Now cover the application program. */
 		if (app_tpnt->dynamic_info[DT_TEXTREL]) {
-			ppnt = (elf_phdr *) auxvt[AT_PHDR].a_un.a_ptr;
+			ppnt = (ElfW(Phdr) *) auxvt[AT_PHDR].a_un.a_ptr;
 			for (i = 0; i < auxvt[AT_PHNUM].a_un.a_val; i++, ppnt++) {
 				if (ppnt->p_type == PT_LOAD && !(ppnt->p_flags & PF_W))
 					_dl_mprotect((void *) (ppnt->p_vaddr & PAGE_ALIGN),
@@ -490,24 +508,20 @@ LD_BOOT(unsigned long args)
 			}
 		}
 	}
-
-	/* Now we can store the debug structure address */
-	dpnt_debug->d_un.d_val = (unsigned long) debug_addr;
 #endif
-
-
+	
 #if defined(__mips__)
-	/*
-	 * For MIPS we have to do stuff to the GOT before we do relocations.
-	 */
+#ifdef __SUPPORT_LD_DEBUG_EARLY__
+	SEND_STDERR("About to do MIPS specific GOT bootstrap\n");
+#endif
+	/* For MIPS we have to do stuff to the GOT before we do relocations.  */
 	PERFORM_BOOTSTRAP_GOT(got);
 #endif
 
-
 	/* OK, now do the relocations.  We do not do a lazy binding here, so
 	   that once we are done, we have considerably more flexibility. */
 #ifdef __SUPPORT_LD_DEBUG_EARLY__
-	SEND_STDERR("About to do library loader relocations.\n");
+	SEND_STDERR("About to do library loader relocations\n");
 #endif
 
 	goof = 0;
@@ -608,11 +622,14 @@ LD_BOOT(unsigned long args)
 	_dl_get_ready_to_run(tpnt, app_tpnt, load_addr, hash_addr, auxvt, envp, debug_addr);
 
 
+	/* Notify the debugger that all objects are now mapped in.  */
+	_dl_debug_addr->r_state = RT_CONSISTENT;
+	_dl_debug_state();
+
 
 	/* OK we are done here.  Turn out the lights, and lock up. */
 	_dl_elf_main = (int (*)(int, char **, char **)) auxvt[AT_ENTRY].a_un.a_fcn;
 
-
 	/*
 	 * Transfer control to the application.
 	 */
@@ -635,7 +652,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 		unsigned long load_addr, unsigned long *hash_addr, Elf32_auxv_t auxvt[AT_EGID + 1], 
 		char **envp, struct r_debug *debug_addr)
 {
-	elf_phdr *ppnt;
+	ElfW(Phdr) *ppnt;
 	char *lpntstr;
 	int i, _dl_secure, goof = 0;
 	struct dyn_elf *rpnt;
@@ -657,40 +674,33 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	tpnt->next = 0;
 	tpnt->libname = 0;
 	tpnt->libtype = program_interpreter;
-	tpnt->loadaddr = (char *) load_addr;
+	tpnt->loadaddr = (ElfW(Addr)) load_addr;
 
-	INIT_GOT(lpnt, tpnt);
-#ifdef __SUPPORT_LD_DEBUG_EARLY__
-	_dl_dprintf(_dl_debug_file, "GOT found at %x\n", lpnt);
+#ifdef ALLOW_ZERO_PLTGOT
+	if (tpnt->dynamic_info[DT_PLTGOT])
 #endif
+	{
+		INIT_GOT(lpnt, tpnt);
+#ifdef __SUPPORT_LD_DEBUG_EARLY__
+		_dl_dprintf(_dl_debug_file, "GOT found at %x\n", lpnt);
+#endif
+	}
+
 	/* OK, this was a big step, now we need to scan all of the user images
 	   and load them properly. */
 
 	{
-		elfhdr *epnt;
-		elf_phdr *myppnt;
+		ElfW(Ehdr) *epnt;
+		ElfW(Phdr) *myppnt;
 		int j;
 
-		epnt = (elfhdr *) auxvt[AT_BASE].a_un.a_ptr;
+		epnt = (ElfW(Ehdr) *) auxvt[AT_BASE].a_un.a_ptr;
 		tpnt->n_phent = epnt->e_phnum;
-		tpnt->ppnt = myppnt = (elf_phdr *) (load_addr + epnt->e_phoff);
+		tpnt->ppnt = myppnt = (ElfW(Phdr) *) (load_addr + epnt->e_phoff);
 		for (j = 0; j < epnt->e_phnum; j++, myppnt++) {
 			if (myppnt->p_type == PT_DYNAMIC) {
-				tpnt->dynamic_addr = myppnt->p_vaddr + load_addr;
-#if defined(__mips__)
-				{
-					int k = 1;
-					Elf32_Dyn *dpnt = (Elf32_Dyn *) tpnt->dynamic_addr;
-
-					while(dpnt->d_tag) {
-						dpnt++;
-						k++;
-					}
-					tpnt->dynamic_size = k * sizeof(Elf32_Dyn);
-				}
-#else
+				tpnt->dynamic_addr = (ElfW(Dyn) *)myppnt->p_vaddr + load_addr;
 				tpnt->dynamic_size = myppnt->p_filesz;
-#endif
 			}
 		}
 	}
@@ -702,7 +712,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	   and figure out which libraries are supposed to be called.  Until
 	   we have this list, we will not be completely ready for dynamic linking */
 
-	ppnt = (elf_phdr *) auxvt[AT_PHDR].a_un.a_ptr;
+	ppnt = (ElfW(Phdr) *) auxvt[AT_PHDR].a_un.a_ptr;
 	for (i = 0; i < auxvt[AT_PHNUM].a_un.a_val; i++, ppnt++) {
 		if (ppnt->p_type == PT_LOAD) {
 			if (ppnt->p_vaddr + ppnt->p_memsz > brk_addr)
@@ -715,28 +725,13 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 				continue;
 #endif
 			/* OK, we have what we need - slip this one into the list. */
-#if defined(__mips__)
-			{
-				int i = 1;
-				Elf32_Dyn *dpnt = (Elf32_Dyn *) tpnt->dynamic_addr;
-
-				while(dpnt->d_tag) {
-					dpnt++;
-					i++;
-				}
-				app_tpnt = _dl_add_elf_hash_table("", 0, 
-					app_tpnt->dynamic_info, ppnt->p_vaddr,
-					(i * sizeof(Elf32_Dyn)));
-			}
-#else
 			app_tpnt = _dl_add_elf_hash_table("", 0, 
 					app_tpnt->dynamic_info, ppnt->p_vaddr, ppnt->p_filesz);
-#endif
 			_dl_loaded_modules->libtype = elf_executable;
-			_dl_loaded_modules->ppnt = (elf_phdr *) auxvt[AT_PHDR].a_un.a_ptr;
+			_dl_loaded_modules->ppnt = (ElfW(Phdr) *) auxvt[AT_PHDR].a_un.a_ptr;
 			_dl_loaded_modules->n_phent = auxvt[AT_PHNUM].a_un.a_val;
 			_dl_symbol_tables = rpnt = (struct dyn_elf *) _dl_malloc(sizeof(struct dyn_elf));
-			_dl_memset(rpnt, 0, sizeof(*rpnt));
+			_dl_memset(rpnt, 0, sizeof(struct dyn_elf));
 			rpnt->dyn = _dl_loaded_modules;
 			app_tpnt->usage_count++;
 			app_tpnt->symbol_scope = _dl_symbol_tables;
@@ -787,7 +782,8 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	/* Now we need to figure out what kind of options are selected.
 	   Note that for SUID programs we ignore the settings in LD_LIBRARY_PATH */
 	{
-		_dl_not_lazy = _dl_getenv("LD_BIND_NOW", envp);
+		if (_dl_getenv("LD_BIND_NOW", envp))
+			_dl_be_lazy = 0;
 
 		if ((auxvt[AT_UID].a_un.a_val == -1 && _dl_suid_ok()) ||
 				(auxvt[AT_UID].a_un.a_val != -1 && 
@@ -859,21 +855,35 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	
 	
 #endif	
-	_dl_trace_loaded_objects = _dl_getenv("__LDSO_LDD_SUPPORT___LOADED_OBJECTS", envp);
+	_dl_trace_loaded_objects = _dl_getenv("LD_TRACE_LOADED_OBJECTS", envp);
 #ifndef __LDSO_LDD_SUPPORT__
 	if (_dl_trace_loaded_objects) {
 		_dl_dprintf(_dl_debug_file, "Use the ldd provided by uClibc\n");
 		_dl_exit(1);
 	}
 #endif
+
+	/*
+	 * OK, fix one more thing - set up debug_addr so it will point
+	 * to our chain.  Later we may need to fill in more fields, but this
+	 * should be enough for now.
+	 */
+	debug_addr->r_map = (struct link_map *) _dl_loaded_modules;
+	debug_addr->r_version = 1;
+	debug_addr->r_ldbase = load_addr;
+	debug_addr->r_brk = (unsigned long) &_dl_debug_state;
+	_dl_debug_addr = debug_addr;
+
+	/* Notify the debugger we are in a consistant state */
+	_dl_debug_addr->r_state = RT_CONSISTENT;
+	_dl_debug_state();
+
 	/* OK, we now have the application in the list, and we have some
 	   basic stuff in place.  Now search through the list for other shared
 	   libraries that should be loaded, and insert them on the list in the
 	   correct order. */
 
-#ifdef USE_CACHE
 	_dl_map_cache();
-#endif
 
 
 	if (_dl_preload) 
@@ -892,30 +902,35 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 			*str2 = '\0';
 			if (!_dl_secure || _dl_strchr(str, '/') == NULL) 
 			{
+				if ((tpnt1 = _dl_check_if_named_library_is_loaded(str))) 
+				{
+					continue;
+				}
+#if defined (__SUPPORT_LD_DEBUG__)
+				if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tfile='%s';  needed by '%s'\n", 
+						str, _dl_progname);
+#endif
 				tpnt1 = _dl_load_shared_library(_dl_secure, &rpnt, NULL, str);
 				if (!tpnt1) {
 #ifdef __LDSO_LDD_SUPPORT__
 					if (_dl_trace_loaded_objects)
 						_dl_dprintf(1, "\t%s => not found\n", str);
-					else {
+					else 
 #endif
-						_dl_dprintf(2, "%s: can't load "
-								"library '%s'\n", _dl_progname, str);
+					{
+						_dl_dprintf(2, "%s: can't load " "library '%s'\n", _dl_progname, str);
 						_dl_exit(15);
-#ifdef __LDSO_LDD_SUPPORT__
 					}
-#endif
 				} else {
 #ifdef __SUPPORT_LD_DEBUG_EARLY__
 					_dl_dprintf(_dl_debug_file, "Loading:\t(%x) %s\n", tpnt1->loadaddr, tpnt1->libname);
 #endif
 #ifdef __LDSO_LDD_SUPPORT__
-					if (_dl_trace_loaded_objects
-							&& tpnt1->usage_count==1) {
+					if (_dl_trace_loaded_objects && tpnt1->usage_count==1) {
 						/* this is a real hack to make ldd not print 
 						 * the library itself when run on a library. */
 						if (_dl_strcmp(_dl_progname, str) != 0)
-							_dl_dprintf(1, "\t%s => %s (0x%x)\n", str, tpnt1->libname, 
+							_dl_dprintf(1, "\t%s => %s (%x)\n", str, tpnt1->libname, 
 									(unsigned) tpnt1->loadaddr);
 					}
 #endif
@@ -971,27 +986,32 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 						c = *cp;
 						*cp = '\0';
 
+						if ((tpnt1 = _dl_check_if_named_library_is_loaded(cp2))) 
+						{
+							continue;
+						}
+#if defined (__SUPPORT_LD_DEBUG__)
+						if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tfile='%s';  needed by '%s'\n", 
+								cp2, _dl_progname);
+#endif
 						tpnt1 = _dl_load_shared_library(0, &rpnt, NULL, cp2);
 						if (!tpnt1) {
 #ifdef __LDSO_LDD_SUPPORT__
 							if (_dl_trace_loaded_objects)
 								_dl_dprintf(1, "\t%s => not found\n", cp2);
-							else {
+							else 
 #endif
-								_dl_dprintf(2, "%s: can't load library '%s'\n", 
-										_dl_progname, cp2);
+							{
+								_dl_dprintf(2, "%s: can't load library '%s'\n", _dl_progname, cp2);
 								_dl_exit(15);
-#ifdef __LDSO_LDD_SUPPORT__
 							}
-#endif
 						} else {
 #ifdef __SUPPORT_LD_DEBUG_EARLY__
 							_dl_dprintf(_dl_debug_file, "Loading:\t(%x) %s\n", tpnt1->loadaddr, tpnt1->libname);
 #endif
 #ifdef __LDSO_LDD_SUPPORT__
-							if (_dl_trace_loaded_objects
-									&& tpnt1->usage_count==1) {
-								_dl_dprintf(1, "\t%s => %s (0x%x)\n", cp2, 
+							if (_dl_trace_loaded_objects && tpnt1->usage_count==1) {
+								_dl_dprintf(1, "\t%s => %s (%x)\n", cp2, 
 										tpnt1->libname, (unsigned) tpnt1->loadaddr);
 							}
 #endif
@@ -1010,63 +1030,46 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	}
 #endif
 
-	for (tcurr = _dl_loaded_modules; tcurr; tcurr = tcurr->next) {
+	for (tcurr = _dl_loaded_modules; tcurr; tcurr = tcurr->next) 
+	{
 		Elf32_Dyn *dpnt;
-		for (dpnt = (Elf32_Dyn *) tcurr->dynamic_addr; dpnt->d_tag;
-				dpnt++) {
-			if (dpnt->d_tag == DT_NEEDED) {
-				lpntstr = tcurr->loadaddr + tcurr->dynamic_info[DT_STRTAB] +
-					dpnt->d_un.d_val;
-				if (_dl_strcmp(lpntstr, "libc.so.6") == 0) {
-					_dl_dprintf(2, "%s: linked against GNU libc!\n", _dl_progname);
-					_dl_exit(150);
-				}
-				if (tpnt && _dl_strcmp(lpntstr, _dl_get_last_path_component(tpnt->libname)) == 0) {
-					struct elf_resolve *ttmp;
+		for (dpnt = (Elf32_Dyn *) tcurr->dynamic_addr; dpnt->d_tag; dpnt++) 
+		{
+			if (dpnt->d_tag == DT_NEEDED) 
+			{
+				char *name;
+				lpntstr = (char*) (tcurr->loadaddr + tcurr->dynamic_info[DT_STRTAB] + dpnt->d_un.d_val);
+				name = _dl_get_last_path_component(lpntstr);
 
-#ifdef __LDSO_LDD_SUPPORT__
-					if (_dl_trace_loaded_objects && tpnt->usage_count==1) {
-						_dl_dprintf(1, "\t%s => %s (0x%x)\n", 
-								lpntstr, tpnt->libname, (unsigned) tpnt->loadaddr);
-					}
-#endif
-					ttmp = _dl_loaded_modules;
-					while (ttmp->next)
-						ttmp = ttmp->next;
-					ttmp->next = tpnt;
-					tpnt->prev = ttmp;
-					tpnt->next = NULL;
-					rpnt->next = (struct dyn_elf *)
-						_dl_malloc(sizeof(struct dyn_elf));
-					_dl_memset(rpnt->next, 0, sizeof(*(rpnt->next)));
-					rpnt->next->prev = rpnt;
-					rpnt = rpnt->next;
-					rpnt->dyn = tpnt;
-					tpnt->usage_count++;
-					tpnt->symbol_scope = _dl_symbol_tables;
-					tpnt = NULL;
+				if ((tpnt1 = _dl_check_if_named_library_is_loaded(name))) 
+				{
 					continue;
 				}
+#if defined (__SUPPORT_LD_DEBUG__)
+				if(_dl_debug) _dl_dprintf(_dl_debug_file, "\tfile='%s';  needed by '%s'\n", 
+						lpntstr, _dl_progname);
+#endif
 				if (!(tpnt1 = _dl_load_shared_library(0, &rpnt, tcurr, lpntstr)))
 				{
 #ifdef __LDSO_LDD_SUPPORT__
-					if (_dl_trace_loaded_objects)
+					if (_dl_trace_loaded_objects) {
 						_dl_dprintf(1, "\t%s => not found\n", lpntstr);
-					else {
+						continue;
+					} else 
 #endif
+					{
 						_dl_dprintf(2, "%s: can't load library '%s'\n", _dl_progname, lpntstr);
 						_dl_exit(16);
-#ifdef __LDSO_LDD_SUPPORT__
 					}
-#endif
 				} else {
 #ifdef __SUPPORT_LD_DEBUG_EARLY__
 					_dl_dprintf(_dl_debug_file, "Loading:\t(%x) %s\n", tpnt1->loadaddr, tpnt1->libname);
 #endif
 #ifdef __LDSO_LDD_SUPPORT__
-					if (_dl_trace_loaded_objects && tpnt1->usage_count==1)
-						_dl_dprintf(1, "\t%s => %s (0x%x)\n", lpntstr, tpnt1->libname, 
+					if (_dl_trace_loaded_objects && tpnt1->usage_count==1) {
+						_dl_dprintf(1, "\t%s => %s (%x)\n", lpntstr, tpnt1->libname, 
 								(unsigned) tpnt1->loadaddr);
+					}
 #endif
 				}
 			}
@@ -1074,18 +1077,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	}
 
 
-#ifdef USE_CACHE
 	_dl_unmap_cache();
-#endif
-	/* ldd uses uses this.  I am not sure how you pick up the other flags */
-#ifdef __LDSO_LDD_SUPPORT__
-	if (_dl_trace_loaded_objects) {
-		char *_dl_warn = 0;
-		_dl_warn = _dl_getenv("LD_WARN", envp);
-		if (!_dl_warn)
-			_dl_exit(0);
-	}
-#endif
 
 	/*
 	 * If the program interpreter is not in the module chain, add it.  This will
@@ -1108,26 +1100,28 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 			tpnt->prev = NULL;
 		}
 		if (rpnt) {
-			rpnt->next =
-				(struct dyn_elf *) _dl_malloc(sizeof(struct dyn_elf));
-			_dl_memset(rpnt->next, 0, sizeof(*(rpnt->next)));
+			rpnt->next = (struct dyn_elf *) _dl_malloc(sizeof(struct dyn_elf));
+			_dl_memset(rpnt->next, 0, sizeof(struct dyn_elf));
 			rpnt->next->prev = rpnt;
 			rpnt = rpnt->next;
 		} else {
 			rpnt = (struct dyn_elf *) _dl_malloc(sizeof(struct dyn_elf));
-			_dl_memset(rpnt, 0, sizeof(*(rpnt->next)));
+			_dl_memset(rpnt, 0, sizeof(struct dyn_elf));
 		}
 		rpnt->dyn = tpnt;
 		tpnt = NULL;
 	}
+
 #ifdef __LDSO_LDD_SUPPORT__
+	/* End of the line for ldd.... */
 	if (_dl_trace_loaded_objects) {
-		_dl_dprintf(1, "\t%s => %s (0x%x)\n", rpnt->dyn->libname + (_dl_strlen(_dl_ldsopath)) + 1, 
+		_dl_dprintf(1, "\t%s => %s (%x)\n", rpnt->dyn->libname + (_dl_strlen(_dl_ldsopath)) + 1, 
 				rpnt->dyn->libname, rpnt->dyn->loadaddr);  
 		_dl_exit(0);
 	}
 #endif
 
+
 #ifdef __mips__
 	/*
 	 * Relocation of the GOT entries for MIPS have to be done
@@ -1144,7 +1138,7 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 	 * Now we go through and look for REL and RELA records that indicate fixups
 	 * to the GOT tables.  We need to do this in reverse order so that COPY
 	 * directives work correctly */
-	goof = _dl_loaded_modules ? _dl_fixup(_dl_loaded_modules) : 0;
+	goof = _dl_loaded_modules ? _dl_fixup(_dl_loaded_modules, _dl_be_lazy) : 0;
 
 
 	/* Some flavors of SVr4 do not generate the R_*_COPY directive,
@@ -1156,10 +1150,6 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 #endif
 	if (_dl_symbol_tables)
 		goof += _dl_copy_fixups(_dl_symbol_tables);
-#ifdef __LDSO_LDD_SUPPORT__
-	if (goof || _dl_trace_loaded_objects)
-		_dl_exit(0);
-#endif
 
 	/* OK, at this point things are pretty much ready to run.  Now we
 	   need to touch up a few items that are required, and then
@@ -1180,10 +1170,10 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 		*_dl_envp = (unsigned long) envp;
 	}
 
-#ifdef DO_MPROTECT_HACKS
+#ifndef FORCE_SHAREABLE_TEXT_SEGMENTS
 	{
 		unsigned int j;
-		elf_phdr *myppnt;
+		ElfW(Phdr) *myppnt;
 
 		/* We had to set the protections of all pages to R/W for dynamic linking.
 		   Set text pages back to R/O */
@@ -1204,20 +1194,9 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 		(intptr_t) _dl_find_hash("on_exit", NULL, NULL, symbolrel);
 #endif
 
-	/*
-	 * OK, fix one more thing - set up the debug_addr structure to point
-	 * to our chain.  Later we may need to fill in more fields, but this
-	 * should be enough for now.
-	 */
-	debug_addr->r_map = (struct link_map *) _dl_loaded_modules;
-	debug_addr->r_version = 1;
-	debug_addr->r_ldbase = load_addr;
-	debug_addr->r_brk = (unsigned long) &_dl_debug_state;
-	_dl_debug_addr = debug_addr;
-	debug_addr->r_state = RT_CONSISTENT;
-	/* This is written in this funny way to keep gcc from inlining the
-	   function call. */
-	((void (*)(void)) debug_addr->r_brk) ();
+	/* Notify the debugger we have added some objects. */
+	_dl_debug_addr->r_state = RT_ADD;
+	_dl_debug_state();
 
 	for (rpnt = _dl_symbol_tables; rpnt!=NULL&& rpnt->next!=NULL; rpnt=rpnt->next)
 	  ;
@@ -1240,15 +1219,17 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
 		tpnt->init_flag |= INIT_FUNCS_CALLED;
 
 		if (tpnt->dynamic_info[DT_INIT]) {
-			_dl_elf_init = (int (*)(void)) (intptr_t) (tpnt->loadaddr + tpnt->dynamic_info[DT_INIT]);
-			  
+			void (*dl_elf_func) (void);
+			dl_elf_func = (void (*)(void)) (intptr_t) (tpnt->loadaddr + tpnt->dynamic_info[DT_INIT]);
 #if defined (__SUPPORT_LD_DEBUG__)
 			if(_dl_debug) _dl_dprintf(_dl_debug_file,"\ncalling init: %s\n\n", tpnt->libname);	
 #endif    
-			(*_dl_elf_init) ();
+			(*dl_elf_func) ();
 		}
 		if (_dl_atexit && tpnt->dynamic_info[DT_FINI]) {
-			(*_dl_atexit) (tpnt->loadaddr + tpnt->dynamic_info[DT_FINI]);
+			void (*dl_elf_func) (void);
+			dl_elf_func = (void (*)(void)) (intptr_t) (tpnt->loadaddr + tpnt->dynamic_info[DT_FINI]);
+			(*_dl_atexit) (dl_elf_func);
 #if defined (__SUPPORT_LD_DEBUG__)
 			if(_dl_debug && _dl_on_exit)
 			{
@@ -1274,57 +1255,65 @@ static void _dl_get_ready_to_run(struct elf_resolve *tpnt, struct elf_resolve *a
  * can set an internal breakpoint on it, so that we are notified when the
  * address mapping is changed in some way.
  */
-void _dl_debug_state()
+void _dl_debug_state(void)
 {
-	return;
 }
 
-int _dl_fixup(struct elf_resolve *tpnt)
+int _dl_fixup(struct elf_resolve *tpnt, int flag)
 {
 	int goof = 0;
 
 	if (tpnt->next)
-		goof += _dl_fixup(tpnt->next);
+		goof += _dl_fixup(tpnt->next, flag);
 #if defined (__SUPPORT_LD_DEBUG__)
 	if(_dl_debug) _dl_dprintf(_dl_debug_file,"\nrelocation processing: %s", tpnt->libname);	
 #endif    
 	
 	if (tpnt->dynamic_info[DT_REL]) {
 #ifdef ELF_USES_RELOCA
-		_dl_dprintf(2, "%s: can't handle REL relocation records\n",
-					_dl_progname);
-		_dl_exit(17);
+#if defined (__SUPPORT_LD_DEBUG__)
+		if(_dl_debug) _dl_dprintf(2, "%s: can't handle REL relocation records\n", _dl_progname);
+#endif    
+		goof++;
+		return goof;
 #else
 		if (tpnt->init_flag & RELOCS_DONE)
 			return goof;
 		tpnt->init_flag |= RELOCS_DONE;
-		goof += _dl_parse_relocation_information(tpnt, tpnt->dynamic_info[DT_REL], 
+		goof += _dl_parse_relocation_information(tpnt, 
+				tpnt->dynamic_info[DT_REL], 
 				tpnt->dynamic_info[DT_RELSZ], 0);
 #endif
 	}
 	if (tpnt->dynamic_info[DT_RELA]) {
-#ifdef ELF_USES_RELOCA
+#ifndef ELF_USES_RELOCA
+#if defined (__SUPPORT_LD_DEBUG__)
+		if(_dl_debug) _dl_dprintf(2, "%s: can't handle RELA relocation records\n", _dl_progname);
+#endif    
+		goof++;
+		return goof;
+#else
 		if (tpnt->init_flag & RELOCS_DONE)
 			return goof;
 		tpnt->init_flag |= RELOCS_DONE;
-		goof += _dl_parse_relocation_information(tpnt, tpnt->dynamic_info[DT_RELA], 
+		goof += _dl_parse_relocation_information(tpnt, 
+				tpnt->dynamic_info[DT_RELA], 
 				tpnt->dynamic_info[DT_RELASZ], 0);
-#else
-		_dl_dprintf(2, "%s: can't handle RELA relocation records\n",
-					_dl_progname);
-		_dl_exit(18);
 #endif
 	}
 	if (tpnt->dynamic_info[DT_JMPREL]) {
 		if (tpnt->init_flag & JMP_RELOCS_DONE)
 			return goof;
 		tpnt->init_flag |= JMP_RELOCS_DONE;
-		if (!_dl_not_lazy || *_dl_not_lazy == 0)
-			_dl_parse_lazy_relocation_information(tpnt, tpnt->dynamic_info[DT_JMPREL], 
+		if (flag & RTLD_LAZY) {
+			_dl_parse_lazy_relocation_information(tpnt, 
+					tpnt->dynamic_info[DT_JMPREL], 
 					tpnt->dynamic_info [DT_PLTRELSZ], 0);
-		else
-			goof += _dl_parse_relocation_information(tpnt, tpnt->dynamic_info[DT_JMPREL], 
+		} else {
+			goof += _dl_parse_relocation_information(tpnt, 
+					tpnt->dynamic_info[DT_JMPREL], 
 					tpnt->dynamic_info[DT_PLTRELSZ], 0);
+		}
 	}
 #if defined (__SUPPORT_LD_DEBUG__)
 	if(_dl_debug) {
@@ -1335,41 +1324,6 @@ int _dl_fixup(struct elf_resolve *tpnt)
 	return goof;
 }
 
-void *_dl_malloc(int size)
-{
-	void *retval;
-
-#if 0
-#ifdef __SUPPORT_LD_DEBUG_EARLY__
-	_dl_dprintf(_dl_debug_file, "malloc: request for %d bytes\n", size);
-#endif
-#endif
-
-	if (_dl_malloc_function)
-		return (*_dl_malloc_function) (size);
-
-	if (_dl_malloc_addr - _dl_mmap_zero + size > 4096) {
-#ifdef __SUPPORT_LD_DEBUG_EARLY__
-		_dl_dprintf(_dl_debug_file, "malloc: mmapping more memory\n");
-#endif
-		_dl_mmap_zero = _dl_malloc_addr = _dl_mmap((void *) 0, size, 
-				PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
-		if (_dl_mmap_check_error(_dl_mmap_zero)) {
-			_dl_dprintf(2, "%s: mmap of a spare page failed!\n", _dl_progname);
-			_dl_exit(20);
-		}
-	}
-	retval = _dl_malloc_addr;
-	_dl_malloc_addr += size;
-
-	/*
-	 * Align memory to 4 byte boundary.  Some platforms require this, others
-	 * simply get better performance.
-	 */
-	_dl_malloc_addr = (char *) (((unsigned long) _dl_malloc_addr + 3) & ~(3));
-	return retval;
-}
-
 char *_dl_getenv(const char *symbol, char **envp)
 {
 	char *pnt;
@@ -1403,88 +1357,5 @@ void _dl_unsetenv(const char *symbol, char **envp)
 	return;
 }
 
-char *_dl_strdup(const char *string)
-{
-	char *retval;
-	int len;
-
-	len = _dl_strlen(string);
-	retval = _dl_malloc(len + 1);
-	_dl_strcpy(retval, string);
-	return retval;
-}
-
-/* Minimal printf which handles only %s, %d, and %x */
-void _dl_dprintf(int fd, const char *fmt, ...)
-{
-	int num;
-	va_list args;
-	char *start, *ptr, *string;
-	char buf[2048];
-
-	start = ptr = buf;
-
-	if (!fmt)
-		return;
-
-	if (_dl_strlen(fmt) >= (sizeof(buf) - 1))
-		_dl_write(fd, "(overflow)\n", 10);
-
-	_dl_strcpy(buf, fmt);
-	va_start(args, fmt);
-
-	while (start) {
-		while (*ptr != '%' && *ptr) {
-			ptr++;
-		}
-
-		if (*ptr == '%') {
-			*ptr++ = '\0';
-			_dl_write(fd, start, _dl_strlen(start));
-
-			switch (*ptr++) {
-			case 's':
-				string = va_arg(args, char *);
-
-				if (!string)
-					_dl_write(fd, "(null)", 6);
-				else
-					_dl_write(fd, string, _dl_strlen(string));
-				break;
-
-			case 'i':
-			case 'd':
-			{
-				char tmp[22];
-				num = va_arg(args, int);
-
-				string = _dl_simple_ltoa(tmp, num);
-				_dl_write(fd, string, _dl_strlen(string));
-				break;
-			}
-			case 'x':
-			case 'X':
-			{
-				char tmp[22];
-				num = va_arg(args, int);
-
-				string = _dl_simple_ltoahex(tmp, num);
-				_dl_write(fd, string, _dl_strlen(string));
-				break;
-			}
-			default:
-				_dl_write(fd, "(null)", 6);
-				break;
-			}
-
-			start = ptr;
-		} else {
-			_dl_write(fd, start, _dl_strlen(start));
-			start = NULL;
-		}
-	}
-	return;
-}
-
 #include "hash.c"
 #include "readelflib1.c"
